/* -*- c++ -*- */
/*
 * Copyright 2005 Free Software Foundation, Inc.
 * 
 * This file is part of GNU Radio
 * 
 * GNU Radio is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * GNU Radio is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with GNU Radio; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <iostream>
#include <string>
#include <fstream>
#include <unistd.h>
#include <stdlib.h>
#include <getopt.h>
#include <boost/scoped_array.hpp>
#include <gr_complex.h>
#include <gr_fxpt_nco.h>
#include "time_series.h"


gr_complex
complex_conj_dotprod(const gr_complex *a, const gr_complex *b, unsigned n)
{
  gr_complex acc = 0;
  for (unsigned i = 0; i < n; i++)
    acc += a[i] * conj(b[i]);

  return acc;
}

/*!
 * \brief frequency translate src by normalized_freq
 *
 * \param dst			destination
 * \param src			source
 * \param n			length of src and dst in samples
 * \param normalized_freq	[-1/2, +1/2]
 */
void
freq_xlate(gr_complex *dst, const gr_complex *src, unsigned n, float normalized_freq)
{
  gr_fxpt_nco	nco;
  nco.set_freq(2 * M_PI * normalized_freq);

  for (unsigned int i = 0; i < n; i++){
    gr_complex	phasor(nco.cos(), nco.sin());
    dst[i] = src[i] * phasor;
    nco.step();
  }
}

inline void
write_float(FILE *output, float x)
{
  fwrite(&x, sizeof(x), 1, output);
}

inline void write_complex(FILE *output,gr_complex x)
{
  fwrite(&x, sizeof(x), 1, output);
}

// write 8-float header
static void
write_header (FILE *output, float ndoppler_bins, float min_doppler, float max_doppler)
{
  write_float(output, ndoppler_bins);
  write_float(output, min_doppler);
  write_float(output, max_doppler);
  write_float(output, 0);
  write_float(output, 0);
  write_float(output, 0);
  write_float(output, 0);
  write_float(output, 0);
}


void
main_loop(FILE *output, time_series &ref_ts, time_series &scat0_ts,
	  unsigned nranges, unsigned correlation_window_size,
	  float min_doppler, float max_doppler, int ndoppler_bins, unsigned int skip=4)
{
  fprintf(stderr, "ndoppler_bins = %10d\n", ndoppler_bins);
  fprintf(stderr, "min_doppler   = %10f\n", min_doppler);
  fprintf(stderr, "max_doppler   = %10f\n", max_doppler);

  // float scale_factor = 1.0/correlation_window_size;	// FIXME, not sure this is right
  float scale_factor = 1.0;				// FIXME, not sure this is right

  boost::scoped_array<gr_complex>  shifted(new gr_complex[correlation_window_size]);

  // gr_complex shifted[correlation_window_size];		// doppler shifted reference

  float doppler_incr = 0;
  if (ndoppler_bins == 1){
    min_doppler = 0;
    max_doppler = 0;
  }
  else
    doppler_incr = (max_doppler - min_doppler) / (ndoppler_bins - 1);

  write_header(output, ndoppler_bins, min_doppler, max_doppler);

  unsigned long long ro = 0;	// reference offset
  unsigned long long so = 0;	// scatter offset 
  //const unsigned int skip = 4;      //skip autocorrelation peak at zero
  for (unsigned int n = 0; n < nranges; n++){
    if (0){
      fprintf(stdout, "n =  %6d\n", n);
      fprintf(stdout, "ro = %6lld\n", ro);
      fprintf(stdout, "so = %6lld\n", so);
    }
    //    const gr_complex *ref = (const gr_complex *) ref_ts.seek(ro, correlation_window_size);//crosscorrelate
    const gr_complex *ref = (const gr_complex *) scat0_ts.seek(ro, correlation_window_size);//autocorrelate
    const gr_complex *scat0 = (const gr_complex *) scat0_ts.seek(so, correlation_window_size);
    if (ref == 0 || scat0 == 0)
      return;

    for (int nd = 0; nd < ndoppler_bins; nd++){
      float fdop = min_doppler + doppler_incr * nd;
      //fprintf(stderr, "fdop = %10g\n", fdop);
      freq_xlate(&shifted[0], ref, correlation_window_size, fdop);	// generated doppler shifted reference

      gr_complex ccor = (n<skip)?gr_complex(0.0,0.0):complex_conj_dotprod(&shifted[0], scat0, correlation_window_size);
      float out = norm(ccor) * scale_factor;

      // fprintf(output, "%12g\n", out);
      write_float(output, out);
    }

    so += 1;
  }
}

void
reconstruct(FILE *output_direct,FILE *output_echo,const char *rx0_filename,const char *rx1_filename,unsigned long long ref_starting_offset)
{
    unsigned long long pos=0;
    time_series ts_rx0(sizeof(gr_complex), rx0_filename, ref_starting_offset, 0);//reading rx0 from start
    const gr_complex *rx0 = (const gr_complex *) ts_rx0.seek(pos, 1);//reading rx0 from start
    time_series ts_rx1(sizeof(gr_complex), rx1_filename, ref_starting_offset, 0);//reading rx1 from start
    const gr_complex *rx1 = (const gr_complex *) ts_rx1.seek(pos, 1);//reading rx1 from start
    gr_complexd cross_corr=gr_complexd(0.0,0.0);
    gr_complexd corr1=gr_complexd(0.0,0.0);//gr_complex(0.0,0.0);
    double amp_rx0=0.0;
    double amp_rx1=0.0;
    int i,skipped=0,ncorr1=0;
    //determine average phase and amplitude differences between rx0 and rx1
    for(i=0;i<ts_rx0.nsamples_available(pos);i++)
    {
      cross_corr+=rx0[i]*conj(rx1[i]);
      //What we actually want is corr+=rx0[i]/rx1[i]; but this fails if rx1[i] is zero;
      double abs_rx0=abs(rx0[i]);
      double abs_rx1=abs(rx1[i]);
      if(abs_rx1<1e-4)
        skipped++;
      else
      {
        corr1+=rx0[i]/rx1[i];
        ncorr1++;
      }
      //auto_corr0+=rx0[i]*conj(rx0[i]);
      //auto_corr1+=rx1[i]*conj(rx0[i]);
      amp_rx0+=abs_rx0;
      amp_rx1+=abs_rx1;
    }
    //int nsamples_used=d_ref.nsamples_available(pos) - skipped;
    assert(ts_rx0.nsamples_available(pos) - skipped == ncorr1);
    corr1=corr1/double(ncorr1);//(gr_complex(corr1.real()/ncorr1,corr1.imag()/ncorr1);
    //cross_corr=conj(cross_corr)/double(ts_rx0.nsamples_available(pos));//negative ranges
    cross_corr=cross_corr/double(ts_rx0.nsamples_available(pos));//negative ranges
    amp_rx0=amp_rx0/ts_rx0.nsamples_available(pos);
    amp_rx1=amp_rx1/ts_rx0.nsamples_available(pos);
    double amp_cross_corr=abs(cross_corr);
    double phase_cross_corr=arg(cross_corr);
    fprintf(stderr,"phase_cross_corr in degrees=%f\n",360.0*phase_cross_corr/(2*M_PI));//should be 90
    fprintf(stderr,"amp_rx1/amp_rx0=%f\n",amp_rx1/amp_rx0);
    fprintf(stderr,"amp_cross-corr=%f\n",amp_cross_corr);
    fprintf(stderr,"factor=%f\n",amp_cross_corr/(amp_rx1/amp_rx0)); 
    fprintf(stderr,"diff=%f\n",amp_cross_corr-(amp_rx1/amp_rx0)); 
    //gr_complexd corr2=(amp_rx1/amp_rx0)*cross_corr/amp_cross_corr;//negative ranges and residual autocorr gr_complex((amp_rx1/amp_rx0)*cross_corr.real()/(amp_cross_corr),(amp_rx1/amp_rx0)*cross_corr.imag()/(amp_cross_corr));
    gr_complexd corr2=(amp_rx0/amp_rx1)*cross_corr/amp_cross_corr;
    //gr_complex corr=(amp_rx1/amp_rx0)*polar(1.0,atan2(cross_corr));//alternative implementation
    for(i=0;i<ts_rx0.nsamples_available(pos);i++)
    {
      gr_complexd rx1_corr=corr2*gr_complexd(rx1[i]);
      //rx1_corr=corr1*rx1[i];
      //rx1_corr=rx1[i]/cross_corr;
      gr_complexd approximate_ref=0.5*(gr_complexd(rx0[i])+rx1_corr);//TODO check this 2*gr_complexd(rx0[i])-rx1_corr;
      gr_complexd approximate_echo=gr_complexd(rx0[i])-rx1_corr;//This has a complicated distorted phase component dependant on both direction of tx as direction of echo('s)
      write_complex(output_direct, gr_complex(approximate_ref));
      write_complex(output_echo, gr_complex(approximate_echo));
    }
//    fseek (d_output,0,SEEK_SET);//reading back output file from start
//    fseek (d_output1,0,SEEK_SET);//reading back second output file from start
//    block_size=
//    for(int i=0;i<d_ref.nsamples_available(pos);i+=block_size)
//    {
//      int  fread (void * buffer, size_t size, size_t count, FILE * stream); 
//    }
}
static void
usage(const char *argv0)
{
  const char *progname;
  const char *t = std::strrchr(argv0, '/');
  if (t != 0)
    progname = t + 1;
  else
    progname = argv0;
    
  fprintf(stderr, "usage: %s [options] rx0_file rx1_file\n", progname);
  fprintf(stderr, "    -p OUTPUT_FILENAME [default=reconstruct_direct.out]\n");
  fprintf(stderr, "    -q OUTPUT_FILENAME [default=reconstruct_echo.out]\n");
  fprintf(stderr, "    -m MIN_RANGE [default=0]\n");
  fprintf(stderr, "    -M MAX_RANGE [default=300]\n");
  fprintf(stderr, "    -w CORRELATION_WINDOW_SIZE [default=2500]\n");
  fprintf(stderr, "    -s NSAMPLES_TO_SKIP [default=0]\n");
  fprintf(stderr, "    -S RANGES_TO_SKIP [default=4]\n");
  fprintf(stderr, "    -d max_doppler (normalized: [0, +1/2)) [default=.0012]\n");
  fprintf(stderr, "    -n ndoppler_bins [default=31]\n");
}

int
main(int argc, char **argv)
{
  int	ch;
  int min_range =    0;
  int max_range =  300;
  int skip=4;
  const char *rx0_filename = 0;
  const char *rx1_filename = 0;
  const char *output_filename_direct = "reconstruct_direct.out";
  const char *output_filename_echo = "reconstruct_echo.out";
  unsigned int correlation_window_size = 2500;
  long long int nsamples_to_skip = 0;
  double max_doppler = 0.0012;
  int ndoppler_bins = 31;


  while ((ch = getopt(argc, argv, "h:p:q:w:s:")) != -1){
    switch (ch){

    case 'w':
      correlation_window_size = strtol(optarg, 0, 0);
      if (correlation_window_size <= 1){
	usage(argv[0]);
	fprintf(stderr, "    correlation_window_size must be >= 1\n");
	exit(1);
      }
      break;

    case 'p':
      output_filename_direct = optarg;
      break;

    case 'q':
      output_filename_echo= optarg;
      break;
      
    case 's':
      nsamples_to_skip = (long long) strtof(optarg, 0);
      if (nsamples_to_skip < 0){
	usage(argv[0]);
	fprintf(stderr, "    nsamples_to_skip must be >= 0\n");
	exit(1);
      }
      break;


    case '?':
    case 'h':
    default:
      usage(argv[0]);
      exit(1);
    }
  } // while getopt

  if (argc - optind != 2){
    usage(argv[0]);
    exit(1);
  }



  rx0_filename = argv[optind++];
  rx1_filename = argv[optind++];

  FILE *output_direct = fopen(output_filename_direct, "wb");
  if (output_direct == 0){
    perror(output_filename_direct);
    exit(1);
  }
  FILE *output_echo = fopen(output_filename_echo, "wb");
  if (output_echo == 0){
    perror(output_filename_echo);
    exit(1);
  }
  unsigned long long ref_starting_offset = 0;
  unsigned long long scatter_starting_offset = 0;
  reconstruct(output_direct,output_echo,rx0_filename,rx1_filename,ref_starting_offset);
/*
  if (min_range < 0){
    ref_starting_offset = -min_range;
    scatter_starting_offset = 0;
  }
  else {
    ref_starting_offset = 0;
    scatter_starting_offset = min_range;
  }

  ref_starting_offset += nsamples_to_skip;
  scatter_starting_offset += nsamples_to_skip;

  try {
    time_series ref(sizeof(gr_complex), ref_filename, ref_starting_offset, 0);
    time_series scat0(sizeof(gr_complex), scatter_filename, scatter_starting_offset, 0);

    main_loop(output, ref, scat0, nranges, correlation_window_size,
	      -max_doppler, max_doppler, ndoppler_bins, (unsigned) skip);
  }
  catch (std::string &s){
    std::cerr << s << std::endl;
    exit(1);
  }
*/
  return 0;
}

